"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var graphql_1 = require("graphql");
var apollo_utilities_1 = require("apollo-utilities");
var lodash_1 = require("lodash");
var lodash_2 = require("lodash");
var lodash_3 = require("lodash");
var lodash_4 = require("lodash");
var lodash_5 = require("lodash");
var schema_1 = require("./schema");
var transformations_1 = require("./transformations");
function graphqlLodash(query, operationName) {
    var pathToArgs = {};
    var reshape = {};
    var queryAST = typeof query === 'string' ? graphql_1.parse(query) : query;
    traverseOperationFields(queryAST, operationName, function (node, resultPath, alias) {
        var args = getLodashDirectiveArgs(node, schema_1.lodashAST());
        if (args === null) {
            return;
        }
        var argsSetPath = resultPath.concat(['@_']);
        var previousArgsValue = lodash_2.get(pathToArgs, argsSetPath, null);
        if (previousArgsValue !== null && !lodash_3.isEqual(previousArgsValue, args)) {
            throw Error("Different \"@_\" args for the \"" + resultPath.join('.') + "\" path");
        }
        lodash_5.set(pathToArgs, argsSetPath, args);
        lodash_5.set(reshape, resultPath.concat(['_@']), alias[resultPath.length - 1]);
    });
    return {
        apply: !!Object.keys(pathToArgs).length,
        query: stripQuery(queryAST),
        transform: function (data) { return applyLodashDirective(pathToArgs, data); },
        queryShape: pathToArgs,
        reshape: reshape,
    };
}
exports.graphqlLodash = graphqlLodash;
function traverseOperationFields(queryAST, operationName, cb) {
    var fragments = {};
    var operationAST = graphql_1.getOperationAST(queryAST, operationName);
    if (!operationAST) {
        throw new Error('Must provide operation name if query contains multiple operations.');
    }
    queryAST.definitions.forEach(function (definition) {
        if (definition.kind === graphql_1.Kind.FRAGMENT_DEFINITION) {
            fragments[definition.name.value] = definition;
        }
    });
    var resultPath = [];
    var alias = [];
    cb(operationAST, resultPath, alias);
    traverse(operationAST);
    function traverse(root) {
        graphql_1.visit(root, {
            enter: function (node) {
                var _a;
                if (node.kind === graphql_1.Kind.FIELD) {
                    resultPath.push((node.alias || node.name).value);
                    alias.push((_a = {}, _a[(node.alias || node.name).value] = node.name.value, _a));
                }
                if (node.kind === graphql_1.Kind.FRAGMENT_SPREAD) {
                    var fragmentName = node.name.value;
                    var fragment = fragments[fragmentName];
                    if (!fragment) {
                        throw Error("Unknown fragment: " + fragmentName);
                    }
                    traverse(fragment);
                }
            },
            leave: function (node) {
                if (node.kind !== graphql_1.Kind.FIELD) {
                    return;
                }
                cb(node, resultPath, alias);
                resultPath.pop();
                alias.pop();
            },
        });
    }
}
function getLodashDirectiveArgs(node, lodashDirectiveDef) {
    var lodashNode = null;
    for (var _i = 0, _a = node.directives; _i < _a.length; _i++) {
        var directive = _a[_i];
        if (directive.name.value !== lodashDirectiveDef.name) {
            continue;
        }
        if (lodashNode) {
            throw Error("Duplicating \"@_\" on the \"" + node.name.value + "\"");
        }
        lodashNode = directive;
    }
    if (lodashNode === null) {
        return null;
    }
    var args = apollo_utilities_1.argumentsObjectFromField(lodashNode, lodashDirectiveDef);
    return normalizeLodashArgs(lodashNode.arguments, args);
}
function normalizeLodashArgs(argNodes, args) {
    if (!argNodes) {
        return args;
    }
    argNodes = lodash_4.keyBy(argNodes, function (argNode) { return argNode.name.value; });
    var orderedArgs = {};
    lodash_1.each(argNodes, function (node, name) {
        var argValue = args[name];
        if (node.value.kind === 'ObjectValue') {
            orderedArgs[name] = normalizeLodashArgs(node.value.fields, argValue);
        }
        else if (node.value.kind === 'ListValue') {
            var nodeValues = node.value.values;
            orderedArgs[name] = [];
            for (var i = 0; i < nodeValues.length; ++i) {
                orderedArgs[name][i] = normalizeLodashArgs(nodeValues[i].fields, argValue[i]);
            }
        }
        else if (node.value.kind === 'EnumValue' && node.value.value === 'none') {
            orderedArgs[name] = undefined;
        }
        else {
            orderedArgs[name] = argValue;
        }
    });
    return orderedArgs;
}
function stripQuery(queryAST) {
    var _a;
    return graphql_1.visit(queryAST, (_a = {},
        _a[graphql_1.Kind.DIRECTIVE] = function (node) {
            if (node.name.value === '_') {
                return null;
            }
        },
        _a));
}
function applyLodashDirective(pathToArgs, data) {
    if (data === null) {
        return null;
    }
    var changedData = applyOnPath(data, pathToArgs);
    return applyLodashArgs([], changedData, pathToArgs['@_']);
}
function applyOnPath(result, pathToArgs) {
    var currentPath = [];
    return traverse(result, pathToArgs);
    function traverse(root, pathRoot) {
        if (Array.isArray(root)) {
            return root.map(function (item) { return traverse(item, pathRoot); });
        }
        if (typeof root === 'object') {
            var changedObject = root ? Object.assign({}, root) : root;
            for (var key in pathRoot) {
                if (key === '@_') {
                    continue;
                }
                currentPath.push(key);
                var changedValue = traverse(lodash_2.get(root, key), lodash_2.get(pathRoot, key));
                var lodashArgs = lodash_2.get(pathRoot, [key, '@_'].join('.'));
                changedValue = applyLodashArgs(currentPath, changedValue, lodashArgs);
                if (typeof changedValue === 'function') {
                    changedValue(changedObject, key);
                }
                else {
                    lodash_5.set(changedObject, key, changedValue);
                }
                currentPath.pop();
            }
            return changedObject;
        }
        else {
            return root;
        }
    }
}
function applyLodashArgs(path, object, args) {
    try {
        return transformations_1.applyTransformations(object, args);
    }
    catch (e) {
        console.log(path);
        throw e;
    }
}
//# sourceMappingURL=gql.js.map